/*
 * Copyright (c) 2020, Peter Abeles. All Rights Reserved.
 *
 * This file is part of BoofCV (http://boofcv.org).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package boofcv.struct.image;

/**
 * <p>
 * Gray scale image with a pixel type of signed 32-bit integer
 * </p>
 *
 * @author Peter Abeles
 */
public class GrayS32 extends GrayI<GrayS32> {

	public int[] data;

	/**
	 * Creates a new gray scale (single band/color) image.
	 *
	 * @param width number of columns in the image.
	 * @param height number of rows in the image.
	 */
	public GrayS32( int width, int height ) {
		super(width, height);
	}

	/** Creates an image with no data declared and the width/height set to zero. */
	public GrayS32() {}

	@Override
	public void unsafe_set( int x, int y, int value ) {
		data[getIndex(x, y)] = value;
	}

	@Override
	public int unsafe_get( int x, int y ) {
		return data[getIndex(x, y)];
	}

	/**
	 * Sets the value of the specified pixel.
	 *
	 * @param x pixel coordinate.
	 * @param y pixel coordinate.
	 * @param value The pixel's new value.
	 */
	@Override
	public void set( int x, int y, int value ) {
		if (!isInBounds(x, y))
			throw new ImageAccessException("Requested pixel is out of bounds: " + x + " " + y);

		data[getIndex(x, y)] = value;
	}

	@Override
	public ImageDataType getDataType() {
		return ImageDataType.S32;
	}

	@Override
	protected Object _getData() {
		return data;
	}

	@Override public void _setData( Object data ) {
		this.data = (int[])data;
	}

	@Override
	public GrayS32 createNew( int imgWidth, int imgHeight ) {
		if (imgWidth == -1 || imgHeight == -1)
			return new GrayS32();
		return new GrayS32(imgWidth, imgHeight);
	}

	@Override
	public void copyCol( int col, int row0, int row1, int offset, Object array ) {
		int[] dst = (int[])array;
		int idxSrc = startIndex + stride*row0 + col;
		int idxDst = offset;
		int end = idxSrc + (row1 - row0)*stride;
		while (idxSrc < end) {
			dst[idxDst++] = data[idxSrc];
			idxSrc += stride;
		}
	}

	public int[] getData() {
		return data;
	}

	public void setData( int[] data ) {
		this.data = data;
	}
}
